(function (global, factory) {
	typeof exports === 'object' && typeof module !== 'undefined' ? factory(exports) :
	typeof define === 'function' && define.amd ? define('plonk', ['exports'], factory) :
	(factory((global.plonk = {})));
}(this, (function (exports) { 'use strict';

// Basic number formatter
// Passes value unaltered if it is a Number, converts to Number if it's a coercible String, or returns default if null, undefined, or NaN.

function toNumber (n, def) {
  if ( def === void 0 ) def = 0;

  if (n === null || typeof n === 'undefined') {
    return def;
  }
  if (typeof n === 'string') {
    n = +n;
  }
  if (isNaN(n)) { return def; }
  if (typeof n !== 'number') {
    return def;
  }
  return n;
}

// Constrains an input `value` to `min...max` range.

function clamp (n, min, max) {
  n = toNumber(n, 0);
  min = toNumber(min, 0);
  max = toNumber(max, 1);

  if (arguments.length <= 2) {
    max = min || 1;
    min = 0;
  }

  return Math.min(Math.max(n, min), max);
}

// Wraps the passed function so that it can only be called only once
// If multiple functions are passed, an array of wrapped functions is returned
// Only one function may be called once, then all subsuquent calls to any of them are ignored

function once () {
  var fns = [], len = arguments.length;
  while ( len-- ) fns[ len ] = arguments[ len ];

  var called = false;

  var ret = fns.map(function (fn) {
    var res = void 0;

    if (typeof fn === 'function') {
      return function () {
        var args = [], len = arguments.length;
        while ( len-- ) args[ len ] = arguments[ len ];

        if (!called) {
          res = fn.apply(void 0, args);
          called = true;
        }
        return res;
      };
    }

    return function () {
      called = true;
    };
  });

  return ret.length > 1 ? ret : ret[0];
}

//
// Super minimal Promise class that also impliments progress/notify
//
// Notes about this implimentation:
//
//   * setTimeout is used instead of a microtask like MutationObserver so that animation rendering is not interrupted by a higher priority task.
//       the behavior of setTimeout is also fairly uniform across environments (node, web workers, etc),
//       whereas microtask-based implimentations are not.
//   * for the sake of speed, and since it's not outlined in the A+ spec, there is _no_ error handling in progress/notify methods!
//

var PENDING = {};
var FULFILLED = {};
var REJECTED = {};

var Promise = function Promise (resolver) {
  if (typeof this !== 'object') {
    return new Promise(resolver);
  }

  this._done = false;
  this._state = PENDING;
  this._value = null;
  this._h = { res: [], rej: [], prog: [] };

  if (typeof resolver === 'function') {
    initializeResolver(this, resolver);
  }
};

Promise.prototype.done = function done () {
    var args = [], len = arguments.length;
    while ( len-- ) args[ len ] = arguments[ len ];

  subscribe.apply(void 0, [ this ].concat( args ));

  if (this._state !== PENDING) {
    publish(this);
  }

  return this;
};

Promise.prototype.then = function then (onResolved, onRejected, onNotify) {
  var p = new Promise();

  this.done(
    createResolve(p, onResolved),
    createReject(p, onRejected),
    onNotify
    );

  return p;
};

Promise.prototype.catch = function catch$1 (onRejected) {
  return this.then(null, onRejected);
};

Promise.prototype.progress = function progress (fn) {
  if (typeof fn === 'function') {
    subscribe(this, null, null, fn);
  }

  return this;
};

function resolve (promise, val) {

  if (promise === val) {
    var err$1 = new TypeError('Cannot resolve a promise with itself');
    return reject(promise, err$1);
  }

  if (val && (typeof val === 'object' || typeof val === 'function')) {
    var then;

    try {
      then = val.then;
    } catch (err) {
      return reject(promise, err);
    }

    if (then && typeof then === 'function') {
      return initializeResolver(promise, then.bind(val));
    }
  }

  fulfill(promise, val);
}

function fulfill (promise, val) {
  if (promise._state !== PENDING) {
    return;
  }

  promise._done = true;
  promise._state = FULFILLED;
  promise._value = val;
}

function reject (promise, val) {
  if (promise._state !== PENDING) {
    return;
  }

  promise._done = true;
  promise._state = REJECTED;
  promise._value = val;
}

function progress (promise, val) {
  if (promise._state !== PENDING) {
    return;
  }

  promise._value = val;
}

function initializeResolver (promise, resolver) {
  try {
    resolver.apply(
      void 0, once(createResolve(promise), createReject(promise)).concat( [createNotify(promise)] )
      );
  } catch (err) {
    reject(promise, err);
    publish(promise);
  }
}

function createResolve (promise, onResolved) {
  return function () {
    var args = [], len = arguments.length;
    while ( len-- ) args[ len ] = arguments[ len ];

    if (promise._done) { return; }

    if (typeof onResolved === 'function') {
      try {
        var res = onResolved.apply(void 0, args);
        resolve(promise, res);
      } catch (err) {
        reject(promise, err);
      }
    } else {
      resolve.apply(void 0, [ promise ].concat( args ));
    }

    if (promise._state !== PENDING) {
      publish(promise);
    }
  };
}

function createReject (promise, onRejected) {
  return function () {
    var args = [], len = arguments.length;
    while ( len-- ) args[ len ] = arguments[ len ];

    if (promise._done) { return; }

    if (typeof onRejected === 'function') {
      try {
        var res = onRejected.apply(void 0, args);
        resolve(promise, res);
      } catch (err) {
        reject(promise, err);
      }
    } else {
      reject.apply(void 0, [ promise ].concat( args ));
    }

    if (promise._state !== PENDING) {
      publish(promise);
    }
  };
}

function createNotify (promise) {
  return function () {
    var args = [], len = arguments.length;
    while ( len-- ) args[ len ] = arguments[ len ];

    if (promise._done) { return; }

    progress.apply(void 0, [ promise ].concat( args ));
    publish(promise);
  };
}

function subscribe (promise, res, rej, prog) {
  if (typeof res === 'function') {
    promise._h.res.push(once(res));
  }

  if (typeof rej === 'function') {
    promise._h.rej.push(once(rej));
  }

  if (typeof prog === 'function') {
    promise._h.prog.push(prog);
  }
}


function publish (promise) {
  var handlers;

  switch (promise._state) {
    case PENDING:
      handlers = promise._h.prog;
      break;
    case FULFILLED:
      handlers = promise._h.res;
      break;
    case REJECTED:
      handlers = promise._h.rej;
      break;
  }

  if (!handlers.length) {
    return;
  }

  var val = promise._value,
        len = handlers.length;

  setTimeout(function () {
    for (var i = 0; i < len; i++) {
      var h = handlers[i];
      h && h(val);
    }
  }, 0);
}

// A basic Deferred class that is used for all promises internally

var Deferred = function Deferred () {
  var this$1 = this;

  this.promise = new Promise(function (resolve$$1, reject$$1, notify) {
    this$1.resolve = function () {
      var args = [], len = arguments.length;
      while ( len-- ) args[ len ] = arguments[ len ];

      return resolve$$1.apply(void 0, args);
    };
    this$1.reject = function () {
      var args = [], len = arguments.length;
      while ( len-- ) args[ len ] = arguments[ len ];

      return reject$$1.apply(void 0, args);
    };
    this$1.notify = function () {
      var args = [], len = arguments.length;
      while ( len-- ) args[ len ] = arguments[ len ];

      return notify.apply(void 0, args);
    };
  });
};

// High resolution timestamp that uses `performance.now()` in the browser, or `process.hrtime()` in Node.
// Provides a Date-based fallback otherwise.

var now;

if (typeof performance !== 'undefined' && ('now' in performance)) {
  now = performanceNow;
} else if (typeof process === 'object' && process.toString() === '[object process]') {
  now = hrtimeNow.bind(null, hrtimeToTimestamp());
} else {
  now = dateNow.bind(null, Date.now());
}

var now$1 = now;

function performanceNow () {
  return performance.now();
}

function hrtimeToTimestamp () {
  var hr = process.hrtime();
  return hr[0] * 1e9 + hr[1];
}

function hrtimeNow (offset) {
  return (hrtimeToTimestamp() - offset) / 1e6;
}

function dateNow (offset) {
  return Date.now() - offset;
}

var SIXTY_FPS = 1000 / 60;

// Generic high-resolution timer class that forms the basis for all other timers

var Timer = function Timer (time, callback) {
  if (arguments.length >= 2) {
    time = Math.max(toNumber(time, SIXTY_FPS), 0);
  } else if (arguments.length === 1) {
    callback = time;
    time = SIXTY_FPS;
  }

  if (typeof callback !== 'function') {
    throw new TypeError('Timer callback must be a function');
  }

  this._asyncHandler = asyncHandler;
  this._callback = callback;

  this._state = {
    elapsed: 0,
    initialTime: time,
    interval: 0,
    isRunning: false,
    iterations: 0,
    offset: 0,
    prev: 0,
    time: time
  };

};

Timer.prototype.run = function run () {
    var this$1 = this;

  if (this._state.isRunning) {
    return this;
  }

  this._state.prev = now$1();
  this._state.isRunning = true;

  var tick = function () {
    var isRunning = tickHandler(this$1._state, this$1._callback);
    if (isRunning) {
      this$1._asyncHandler(tick);
    }
  };

  this._asyncHandler(tick);

  return this;
};

Timer.prototype.stop = function stop () {
  var elapsed = this._state.elapsed;

  this._state.isRunning = false;
  this.reset();

  return elapsed;
};

Timer.prototype.reset = function reset () {

  this._state.elapsed = 0;
  this._state.interval = 0;
  this._state.iterations = 0;
  this._state.prev = 0;
  this._state.time = this._state.initialTime;

  return this;
};

Timer.prototype.setTime = function setTime (time) {
    if ( time === void 0 ) time = this._state.time;

  this._state.time = this._state.initialTime = Math.max(toNumber(time, this._state.time), 0);

  return this;
};

function asyncHandler (callback) {
  setTimeout(callback, 0);
}

function tickHandler (state, callback) {
  if (!state.isRunning) {
    return state.isRunning;
  }

  // first tick
  if (state.iterations === 0) {
    callback(0, 0, 0);

    state.prev = now$1();
    state.iterations = 1;

    return state.isRunning;
  }

  state.interval = now$1() - state.prev;

  // interval is below target interval
  if (state.interval <= state.time + state.offset) {
    return state.isRunning;
  }

  state.elapsed += state.interval;

  callback(state.interval, state.iterations, state.elapsed);

  if (state.isRunning) {
    state.prev = now$1();
    state.iterations += 1;
  }

  return state.isRunning;

}

// flattens and (potentially) optimizes try/catch of a function call with rest args

function tryFn (fn) {
  var args = [], len = arguments.length - 1;
  while ( len-- > 0 ) args[ len ] = arguments[ len + 1 ];

  if (typeof fn !== 'function') {
    throw new TypeError('tryFn requires a function');
  }

  var err, res;

  try {
    res = fn.apply(void 0, args);
  } catch (e) {
    err = e;
  }

  return [err, res];
}

// A timer loop, similar to setInterval, except that the time can be reset by the return value of `callback`

function delay (time, callback) {
  if (typeof callback !== 'function') {
    throw new TypeError('delay callback needs to be a function');
  }

  var def = new Deferred();

  var timer = new Timer(time, function (interval, iterations, elapsed) {
    var ref = tryFn(
      callback,
      interval,
      iterations,
      elapsed,
      stop
      );
    var err = ref[0];
    var res = ref[1];

    if (err) {
      timer.stop();
      return def.reject(err);
    }

    timer.setTime(res);
    def.notify(interval);
  });

  timer.run();

  function stop () {
    var elapsed = timer.stop();
    def.resolve(elapsed);
  }

  return def.promise;
}

// Returns a random number in `min...max` range.

function rand (min, max) {
  if (arguments.length <= 1) {
    max = toNumber(min, 1);
    min = 0;
  } else {
    min = toNumber(min, 0);
    max = toNumber(max, 1);
  }

  return Math.random() * (max - min) + min;
}

// Factory that returns a drunk walk random function that walks between `min...max`

function drunk (min, max, step) {
  min = toNumber(min, 0);
  max = toNumber(max, 1);
  step = clamp(toNumber(step, 0.1), 0, 1);

  if (arguments.length <= 1) {
    max = min || 1;
    min = 0;
  }

  var n = rand(min, max);

  return function (s) {
    step = toNumber(s, step);
    n = clamp(n + (max * rand(-1, 1) * step), min, max);
    return n;
  };
}

// Timer function where the tick interval jitters between `min...max` milliseconds.

function dust (min, max, callback) {
  if (typeof callback !== 'function') {
    throw new TypeError('delay callback needs to be a function');
  }

  min = toNumber(min, 10);
  max = toNumber(max, 100);

  if (arguments.length <= 2) {
    max = min;
    min = 0;
  }

  return delay(rand(min, max), function () {
    var args = [], len = arguments.length;
    while ( len-- ) args[ len ] = arguments[ len ];

    callback.apply(void 0, args);

    return rand(min, max);
  });
}

// A repeating timer loop (like setInterval) where `time` is the tick interval in milliseconds.

function metro (time, callback) {
  if (typeof callback !== 'function') {
    throw new TypeError('metro callback needs to be a function');
  }

  var def = new Deferred();

  var timer = new Timer(time, function (interval, iterations, elapsed) {
    var ref = tryFn(
      callback,
      interval,
      iterations,
      elapsed,
      stop
      );
    var err = ref[0];
    var res = ref[1];

    if (err) {
      timer.stop();
      return def.reject(err);
    }

    def.notify(res);
  });

  timer.run();

  function stop (val) {
    timer.stop();
    def.resolve(val);
  }

  return def.promise;
}

// Linear map of `value` in `minIn...maxIn` range to `minOut...maxOut` range.

function scale (n, a1, a2, b1, b2) {
  n = toNumber(n, 0);
  a1 = toNumber(a1, 0);
  a2 = toNumber(a2, 1);
  b1 = toNumber(b1, 0);
  b2 = toNumber(b2, 1);

  return b1 + (clamp(n, a1, a2) - a1) * (b2 - b1) / (a2 - a1);
}

// An envelope that provides linear interpolation of `value` to `target` over `time`

function env (value, target, time, callback) {
  value = toNumber(value, 0);
  target = toNumber(target, 1);
  time = toNumber(time, 100);

  return metro(1000 / 60, function (interval, i, elapsed, stop) {
    var interpolated = scale(elapsed, 0, time, value, target);

    if (typeof callback === 'function') {
      callback(interpolated, elapsed, stop);
    }

    if (elapsed >= time) {
      stop(interpolated);
    }

    return interpolated;
  });

}

// An exponential map of `value` in `0...1` range by Euler's number

function exp (n) {
  n = toNumber(n, 0);
  return Math.pow(clamp(n, 0, 1), Math.E);
}

var animationFrame;

if (typeof window === 'object') {
  var availableFrames = [
    window.requestAnimationFrame,
    window.webkitRequestAnimationFrame,
    window.mozRequestAnimationFrame,
    window.msRequestAnimationFrame
  ];

  for (var i = 0; i < availableFrames.length; i++) {
    var frame = availableFrames[i];
    if (typeof frame === 'function') {
      animationFrame = frame.bind(window);
      break;
    }
  }
}

if (!animationFrame) {
  animationFrame = timeoutAnimationFrame;
}

var animationFrame$1 = animationFrame;

function timeoutAnimationFrame (callback) {
  if (typeof callback !== 'function') {
    throw new TypeError('Not enough arguments');
  }
  setTimeout(callback, 0);
}

// just modifies Timer to use requestAnimtionFrame instead of setTimeout

var Frames = (function (Timer$$1) {
  function Frames () {
    var args = [], len = arguments.length;
    while ( len-- ) args[ len ] = arguments[ len ];

    Timer$$1.apply(this, args);

    this._asyncHandler = animationFrame$1;
    this._state.offset = -5;
  }

  if ( Timer$$1 ) Frames.__proto__ = Timer$$1;
  Frames.prototype = Object.create( Timer$$1 && Timer$$1.prototype );
  Frames.prototype.constructor = Frames;

  return Frames;
}(Timer));

// Animation loop and requestAnimationFrame polyfill with a little extra sugar

function frames (frameRate, callback) {
  if (arguments.length >= 2) {
    frameRate = clamp(toNumber(frameRate, 60), 1, 60);
  } else if (arguments.length === 1) {
    callback = frameRate || callback;
    frameRate = 60;
  }

  if (typeof callback !== 'function') {
    throw new TypeError('frames callback needs to be a function');
  }

  var def = new Deferred();

  var timer = new Frames(1000 / frameRate, function (interval, iterations, elapsed) {
    var ref = tryFn(
      callback,
      interval,
      iterations,
      elapsed,
      stop
      );
    var err = ref[0];
    var res = ref[1];

    if (err) {
      timer.stop();
      return def.reject(err);
    }

    if (typeof res === 'number') {
      frameRate = clamp(toNumber(res, frameRate), 0, 60);

      if (frameRate === 0) {
        return stop();
      }

      timer.setTime(1000 / frameRate);
    }

    def.notify(interval);
  });

  timer.run();

  function stop () {
    var elapsed = timer.stop();
    def.resolve(elapsed);
  }

  return def.promise;
}

var FORMAT_IDENTIFIERS = ['fps', 'hz', 'ms', 's', 'm'];

// Number format converter that takes a variety of input time values and returns the equivalent millisecond values.

function toMilliseconds (val, format, def) {
  if ( format === void 0 ) format = 'ms';
  if ( def === void 0 ) def = 0;

  var ms = toNumber(def, 0);

  if (typeof val === 'string') {
    val = val.toLowerCase();
    for (var i = 0; i < FORMAT_IDENTIFIERS.length; i++) {
      if (val.indexOf(FORMAT_IDENTIFIERS[i]) !== -1) {
        format = FORMAT_IDENTIFIERS[i];
        val = val.replace(' ', '').replace(format, '');
        break;
      }
    }
    val = +val;
  }

  if (val === null || typeof val === 'undefined') {
    return ms;
  }

  if (isNaN(val)) { return ms; }

  switch (format) {
    case 'fps':
      ms = 1000 / val;
      break;
    case 'hz':
      ms = (1 / val) * 1000;
      break;
    case 'ms':
      ms = val;
      break;
    case 's':
      ms = val * 1000;
      break;
    case 'm':
      ms = val * 60 * 1000;
      break;
    default:
  }

  return ms;
}

var SINE_PERIOD = (Math.PI * 2) - 0.0001;

// A sine LFO where `period` is the time, in milliseconds, of one full cycle.

function sine (time, callback) {
  if (typeof callback !== 'function') {
    throw new TypeError('sine callback needs to be a function');
  }

  time = toNumber(time, 0);

  var cycle = 0;

  return metro(1000 / 60, function (interval, i, elapsed, stop) {
    if (cycle >= time) {
      cycle = 0;
    } else {
      cycle += interval;
    }

    var rad = scale(cycle, 0, time, 0, SINE_PERIOD),
          sin = clamp(Math.sin(rad), -1, 1);

    var progress = callback(sin, cycle, elapsed, stop);
    time = toNumber(progress, time);

    return sin;
  });
}

// A simple wrapper for setTimeout that returns a promise.

function wait (time, callback) {
  time = toNumber(time, 0);

  var def = new Deferred(),
        start = now$1();

  setTimeout(function () {
    var elapsed = now$1() - start;

    if (typeof callback === 'function') {
      callback(elapsed);
    }

    def.resolve(elapsed);
  }, time);

  return def.promise;
}

// Timer function where the tick interval performs a drunk walk between `min...max` milliseconds

function walk (min, max, callback) {
  min = toNumber(min, 10);
  max = toNumber(max, 100);

  if (arguments.length <= 2) {
    max = min;
    min = 0;
  }

  var d = drunk(min, max);

  return delay(d(), function () {
    var args = [], len = arguments.length;
    while ( len-- ) args[ len ] = arguments[ len ];

    var progress;

    if (typeof callback === 'function') {
      progress = callback.apply(void 0, args);
    }

    return d(progress);
  });
}

exports.clamp = clamp;
exports.Deferred = Deferred;
exports.delay = delay;
exports.drunk = drunk;
exports.dust = dust;
exports.env = env;
exports.exp = exp;
exports.frames = frames;
exports.metro = metro;
exports.ms = toMilliseconds;
exports.now = now$1;
exports.rand = rand;
exports.scale = scale;
exports.sine = sine;
exports.Timer = Timer;
exports.toMilliseconds = toMilliseconds;
exports.toNumber = toNumber;
exports.wait = wait;
exports.walk = walk;

Object.defineProperty(exports, '__esModule', { value: true });

})));
